#include "mempool.h"
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <errno.h>

mempool *mempool_create(char *buffer, long size)
{
	mempool *mp;
	mp_block_header h;

	/* map memory pool structure to buffer*/
	mp = (mempool *) buffer;
	mp->index = 0;
	mp->size = size - sizeof(mempool);
	
	/* map the remainder to the pool pointer*/
	mp->pool = &buffer[sizeof(mempool)];

	/* create one block of free memory */
	h.currupt = 0;
	h.inuse = 0;
	h.size = size - sizeof(mempool) - sizeof(mp_block_header);

	if (h.size < 1) {
		fprintf(stderr,"%s at line %d, Unable to allocate a memory pool of only %ld bytes, must be at least %d\n",
			__FILE__,__LINE__,size, sizeof(mempool) + sizeof(mp_block_header) + 1);
		return NULL;
		}

	memset( mp->pool, 0, mp->size);
	memcpy( &mp->pool[0], &h, sizeof(mp_block_header));
	
	/* return pointer to pool (same as buffer) */
	return mp;
}


/* seek to a free block large enough for bytes, if found returns 0 else -1 on failure*/
int mempool_seek(mempool *self, long bytes)
{
	long count = 0;
	mp_block_header *h;

	while (count < self->size) {
		int n;

		h = (mp_block_header *) &self->pool[self->index];
		/* curruption check */
		if (h->currupt) {
			fprintf(stderr,"Memory curruption detected in mempool_seek dumping core ... \n");
			abort();
			return -1;
			}

		/* is it big enough and not in use ?? */	
		if (h->inuse == 0 && h->size >= bytes) {
			return 0; /* found it */
			}		

		/* continue to seek */	
		n =  h->size + sizeof(mp_block_header);
		self->index += n;
		count += n;

		if (self->index >= self->size)
			self->index = 0;

		}

	self->index = 0;
              return -1;
}



char *mempool_block_partition(mempool *self, int bytes)
{
	mp_block_header *h, *h2;
	int segsize;

	h = (mp_block_header *) &self->pool[self->index];

	/* should we partition ?? */
	segsize = h->size - bytes - sizeof(mp_block_header);
	if (segsize > 0) {
		/* first seek ahead, create a new block */

		/* will be go off the edge of the pool ? */
		if ((sizeof(mp_block_header) + bytes + self->index) >= self->size) {
			h->currupt = 0;
			h->inuse = 1;	
			h->size = self->size - self->index - sizeof(mp_block_header);
			return ((char *) h) + sizeof(mp_block_header) ;
			}


		self->index += (sizeof(mp_block_header) + bytes);

		h2 = (mp_block_header *) &self->pool[self->index];
		h2->currupt = 0;
		h2->inuse = 0;
		h2->size = segsize;

		/* now checkout the first block */
		h->currupt = 0;
		h->inuse = 1;
		h->size = bytes;

		return ((char *) h) + sizeof(mp_block_header) ;
		}

              /* no we cant partition, so just check out block, or we will go out
                  of bounds */
	h->currupt = 0;
	h->inuse = 1;	
	/* keep the size the same */
	return ((char *) h) + sizeof(mp_block_header) ;
}


void join_seg(mp_block_header *f, mp_block_header *h)
{
	f->size += (h->size + sizeof(mp_block_header));
	h->currupt = 1; /* this header is no longer valid  */
}

/* join adjacent blocks of free memory
*/
void mempool_defrag(mempool *self)
{
	mp_block_header *h, *f = NULL;

	for (self->index = 0; self->index < self->size; ) {
		h = (mp_block_header *) &self->pool[ self->index ];
		if (h->inuse == 0) {

			if (f == NULL)
				f = h; 		/* initailize f */
			else
				join_seg(f,h);   	/* join adjacent to f*/	

			} else {
                                         	f = NULL;
			}	
		/* advance index */
		self->index += (h->size + sizeof(mp_block_header));
		}
	/* reset index */
	self->index = 0;
}


char *mempool_malloc(mempool *self, long bytes)
{
	/* seek to free block*/
	if (mempool_seek(self, bytes)) {
		/* try defragging before we die */
		mempool_defrag(self);
		/* now try again */
		if (mempool_seek(self, bytes)) {
			fprintf(stderr,"Memory pool exhausted\n");
			errno = ENOMEM;
			return NULL;
			}
		}

	/* partition the free block, if possible */
	return mempool_block_partition(self, bytes);
}



void mempool_free(mempool *self,  char *ptr)
{
	mp_block_header *h;

	/* go back and get the header */	
	((char *) h) = ptr - sizeof(mp_block_header);

	/* error check */
	if (h->currupt) {
		fprintf(stderr,"Memory curruption detected in mempool_free dumping core ... \n");
		abort();
		return;
		}

	/* mark as free */
	h->inuse = 0;	
}

void mempool_dump(mempool *self)
{
	long i;
	long next_header=0;

	printf("Memory pool dump\n");
	printf("size=%ld, index = %ld\n", self->size, self->index);

	for (i = 0; i < self->size; ) {
		if (next_header == 0) {
			mp_block_header *h = (mp_block_header *) &self->pool[i];
                                         printf("%04ld [%s, %s, %d]\n", i,(h->currupt) ? "currupt":"ok", (h->inuse)?"inuse":"free", h->size);
			next_header = (long) h->size;
			i += sizeof(mp_block_header);			
			} else {
			printf("%04ld (%X,%c)\n",i,self->pool[i],self->pool[i]);
			i++;
			next_header--;
			}
		}	
}


/* returns how much memory is allocated to this pointer or -1 if the address is invalid
   or no memory assigned to it */
long mempool_getsize(mempool *self, char *ptr)
{
	mp_block_header *h;

	/* see if pointer is in the range of the memory pool */
	if (ptr >= self->pool  && ptr <= &self->pool[self->size]) {
		/* go back and get the header */	
		((char *) h) = ptr - sizeof(mp_block_header);

		/* if no currupt and inuse then valid, return size */
		if ( (!h->currupt) && h->inuse)
			return h->size;
		}

	return -1;
}


/* parities on calloc and realloc */
char *mempool_realloc(mempool *self, char *ptr, long new_size)
{
	char *nptr;

	/* validate pointer  */
	if (mempool_getsize(self,ptr) == -1) {
                           errno = ENOMEM;
		return NULL;
		}

	nptr = mempool_malloc(self,new_size);		
	memcpy(nptr,ptr,new_size);
	mempool_free(self, ptr);
	return nptr;
}


char *mempool_calloc(mempool *self, long size)
{
	char *ptr = mempool_malloc(self,size);
	if (ptr == NULL) return NULL;
	memset(ptr,0,size);
	return ptr;	
}














#ifdef TEST
/* test harness */


#define BUFSIZE 80

int main(int argc, char *argv[])
{
	char buffer[BUFSIZE];
	mempool *mp;
	char *ptr[5];
	int i;
	

	mp = mempool_create(buffer,BUFSIZE);

	for (i = 0; i < 5; i++) {
		if ((ptr[i] = mempool_malloc(mp,10)) == NULL) {
                                   	printf("No memory for %d\n",i);
			return 0;
			}
		
		sprintf(ptr[i],"testing%d",i);
		printf(ptr[i]);
		}

	mempool_free(mp, ptr[2]);
	mempool_free(mp, ptr[3]);
	
	ptr[2] = mempool_malloc(mp,15);


	ptr[2] = mempool_realloc(mp,ptr[2],20);
	sprintf(ptr[2],"1234567890123456789");

	mempool_dump(mp);

	return 0;
}
#endif